/*
 * assembler.c: Read the VM-R Assembly code specification of a program
 *              and produce the executable code in binary or hexadecimal
 *              form.
 *              The syntax of the VM-R Assembly language is the following.
 *              Keywords are in Capital letters only.
 *              Identifiers start with lower case letters only.
 *              Integers are coded in decimal form.
 *              Spaces, tabs and end-of-line can be inserted in arbitrary way.
 *              Single line comments start with a "#" sign and end with the
 *              current line; they may contain any sequence of ASCII
 *              characters except end-of-line.
 *              Multiple line comments start with a "/*" string and end with
 *              a string with the two characters in reverse order ("*" first
 *              followed by "/" with no intermediate space).
 *           
 *     <program>     ::= <instruct> <program>
 *     <program>     ::= <instruct> .
 *     <instruct>    ::= <label> : <instru>
 *     <instruct>    ::= <label> : <vlist> ;
 *     <vlist>       ::= <val16> , <vlist>
 *     <vlist>       ::= <val16>
 *     <instruct>    ::= <vlist> ;
 *     <instruct>    ::= <instru>
 *     <label>       ::= <ident>
 *     <label>       ::= <int16>
 *     <instru>      ::= <instrF1>
 *     <instru>      ::= <instrF2>
 *     <instru>      ::= <instrF3>
 *     <instru>      ::= <instrFJ>
 *     <instru>      ::= <instrps>
 *     <instru>      ::= <instrld>
 *     <instrF3>     ::= ADD3 <reg>, <reg>, <reg>
 *     <instrF3>     ::= AND3 <reg>, <reg>, <reg>
 *     <instrF3>     ::= LOR3 <reg>, <reg>, <reg>
 *     <instrF3>     ::= LOAD <reg>, <reg>, <reg>
 *     <instrF3>     ::= SUB3 <reg>, <reg>, <reg>
 *     <instrF3>     ::= NAND <reg>, <reg>, <reg>
 *     <instrF3>     ::= MOVR <reg>, <reg>, <reg>
 *     <instrF3>     ::= STOR <reg>, <reg>, <reg>
 *     <instrF2>     ::= ROTA <reg>, <reg>, <sconst4>
 *     <instrF2>     ::= PUSH <reg>, <reg>, <uconst4>
 *     <instrF2>     ::= SHFT <reg>, <reg>, <sconst4>
 *     <instrF2>     ::= POPR <reg>, <reg>, <uconst4>
 *     <instrF1>     ::= ADD1 <reg>, <sconst8>
 *     <instrF1>     ::= AND1 <reg>, <uconst8>
 *     <instrF1>     ::= LDIB <reg>, <sconst8>
 *     <instrFJ>     ::= CJMP <jcond> <reladdr>
 *     <reladdr>     ::= <sconst9>
 *     <reladdr>     ::= <ident>
 *     <instrps>     ::= MOV2 <reg>, <reg>
 *     <instrps>     ::= NOT2 <reg>, <reg>
 *     <instrps>     ::= SCMP <reg>, <reg>
 *     <instrps>     ::= LDIW <reg>, <label>
 *     <instrps>     ::= HALT
 *     <instrps>     ::= JUMP <reladdr>
 *     <instrps>     ::= CALL <reg>
 *     <instrps>     ::= RETN
 *     <instrld>     ::= DEFL <ident> = <int16>
 *
 * Copyright C 2001 by Giovanni Chiola <chiola@disi.unige.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Release: 1.0, 11/Jan/2001
 *
 */

#define MAX_DEF 1000
#define MAX_CHAR 1000

#include <stdlib.h>
#include <stdio.h>
#ifdef MAC
#	include <strings.h>
#endif MAC

#ifndef TRUE
#  define TRUE 1
#  define FALSE 0
#endif TRUE

#define opcode_ADD3 0x0000
#define opcode_AND3 0x1000
#define opcode_LOR3 0x2000
#define opcode_LOAD 0x3000
#define opcode_SUB3 0x4000
#define opcode_NAND 0x5000
#define opcode_MOVR 0x6000
#define opcode_STOR 0x7000
#define opcode_ROTA 0x8000
#define opcode_PUSH 0x9000
#define opcode_SHFT 0xA000
#define opcode_POPR 0xB000
#define opcode_ADD1 0xC000
#define opcode_AND1 0xD000
#define opcode_LDIB 0xE000
#define opcode_CJMP 0xF000

#define opcode_MOV2 0x10000
#define opcode_NOT2 0x25000
#define opcode_SCMP 0x34000
#define opcode_LDIW 0x4B040
#define opcode_JUMP 0x5F000
#define opcode_HALT 0x6F1FF
#define opcode_CALL 0x76740
#define opcode_RETN 0x86047

#define Identif     -1
#define Number      -2
#define EndLabel    -3
#define EndList     -4
#define Assign      -5
#define InstrSepar  -6
#define DefDecl     -7
#define EndProgr    -8
#define DefCmpAL    -9
#define DefCmpEQ   -10
#define DefCmpGE   -11
#define DefCmpGT   -12
#define DefCmpLE   -13
#define DefCmpLT   -14
#define DefCmpNE   -15
#define DefCmpOV   -16
#define IsReg(r)   (((r) < -99)&&((r) > -116))
#define NoReg(r)   (-100-(r))

#define VMR_RAM_Size 32768

typedef struct def_str {
  char * nam;
  unsigned val;
} Def_label;

typedef struct und_str {
  char * nam;
  int fst;
} UnD_label;

static unsigned RAM[VMR_RAM_Size];
static int undef_link[VMR_RAM_Size];
static unsigned undef_mask[VMR_RAM_Size];
static int undef_add[VMR_RAM_Size];
static char * Rlabels[VMR_RAM_Size];
static unsigned char cell_used[VMR_RAM_Size];

static Def_label identifier[MAX_DEF];
static UnD_label unidentif[MAX_DEF];
static unsigned no_ident;
static unsigned no_unid;

static unsigned cur_pc;
static unsigned max_pc;

void init(void){
  int i;
  for ( i = 0 ; i < VMR_RAM_Size ; i++ ) {
      RAM[i] = 0; cell_used[i] = FALSE;
      Rlabels[i] = NULL;
      undef_mask[i] = 0;
      undef_add[i] = 0;
      undef_link[i] = -1;
    }
  no_ident = no_unid = 0;
  cur_pc = max_pc = 0;
}

char * my_str_alloc(const char * s) {
  char * my_str;
  size_t len;
  if ( (my_str = (char *)malloc(len = strlen(s)+1)) == NULL ) {
      fprintf(stdout,"\nmy_str_alloc ERROR: couldnt allocate %1d bytes !\n\n",
              (int)len);
      exit(1);
    }
  return strcpy(my_str,s);
}

static char my_string[MAX_CHAR];
static int my_number;

inline void syntax_error(char * my_err) {
    fprintf(stdout,"\nSYNTAX ERROR: %s ! Exiting\n", my_err);
    exit(1);
}

static int next_lex_term(FILE * in_f, FILE * out_f) {
    char nc, c2, c3, pc;
    int i;

      pc = ' ';
      while ( ((nc = getc(in_f)) == '#') || (nc == ' ') || (nc == '\t') ||
              (nc == '\n') || (nc == '/') || ((nc == '*')&&(pc == '/')) ) {
          putc(nc,out_f);
          pc = nc;
          if ( nc == '#' ) {
              while ( (c2 = getc(in_f)) != '\n' )
                  putc(c2,out_f);
              putc('\n',out_f);
            }
          else if ( nc == '*' ) {
              for ( pc = ' ', c2 = getc(in_f); (pc  != '*')&&(c2 != '/') ;
                    pc = c2, c2 = getc(in_f) ) {
                  putc(c2,out_f);
                }
              putc('/',out_f);
            }
        }
      if ( nc >= 'a' && nc <= 'z' ) /* Identifier */ {
          putc(nc,out_f);
          my_string[0] = nc; i = 1;
          while ( (((c2 = getc(in_f)) >= 'a') && (c2 <= 'z'))
                  || (c2 == '_') || ((c2 >= '0') && (c2 <= '9'))
                  || ((c2 >= 'A') && (c2 <= 'Z')) ) {
              my_string[i] = c2; i++;
              putc(c2,out_f);
            }
          ungetc(c2,in_f); my_string[i] = '\0';
          return Identif;
      } else if ( nc >= '0' && nc <= '9' ) /* Number */ {
          ungetc(nc,in_f);
          fscanf(in_f,"%d",&my_number);
          fprintf(out_f,"%1d",my_number);
          return Number;
      } else switch (nc) {
        case '-': ungetc(nc,in_f);
                  fscanf(in_f,"%d",&my_number);
                  fprintf(out_f,"%1d",my_number);
                  return Number;
                 break;
        case ':': putc(nc,out_f);
                  return EndLabel;
                 break;
        case ';': putc(nc,out_f);
                  return EndList;
                 break;
        case '=': putc(nc,out_f);
                  return Assign;
                 break;
        case ',': putc(nc,out_f);
                  return InstrSepar;
                 break;
        case '.': putc(nc,out_f);
                  return EndProgr;
                 break;
        case 'A': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'D' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'D' )
                          syntax_error("D expected after AD");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 == '3' )
                          return opcode_ADD3;
                      if ( c2 == '1' )
                          return opcode_ADD1;
                      syntax_error("1 or 3 expected after ADD");
                  } else if ( c2 == 'L' ) {
                      return DefCmpAL;
                  } else if ( c2 == 'N' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'D' )
                          syntax_error("D expected after AN");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 == '3' )
                          return opcode_AND3;
                      if ( c2 == '1' )
                          return opcode_AND1;
                      syntax_error("1 or 3 expected after AND");
                  } else
                      syntax_error("D or L or N expected after A");
                 break;
        case 'C': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'A' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'L' )
                          syntax_error("L expected after CA");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'L' )
                          syntax_error("L expected after CAL");
                      return opcode_CALL;
                  } else if ( c2 == 'J' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'M' )
                          syntax_error("M expected after CJ");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'P' )
                          syntax_error("P expected after CJM");
                      return opcode_CJMP;
                  } else 
                      syntax_error("A or J expected after C");
                 break;
        case 'D': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'E' )
                      syntax_error("E expected after D");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'F' )
                      syntax_error("F expected after DE");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'L' )
                      syntax_error("L expected after DEF");
                  return DefDecl;
                 break;
        case 'E': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'Q' ) {
                      return DefCmpEQ;
                  } else 
                      syntax_error("Q expected after E");
                 break;
        case 'F': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'P' ) {
                      return NoReg(6);
                  } else 
                      syntax_error("P expected after F");
                 break;
        case 'G': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'E' ) {
                      return DefCmpGE;
                  } else if ( c2 == 'T' ) {
                      return DefCmpGT;
                  } else 
                      syntax_error("E or T expected after G");
                 break;
        case 'H': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'A' )
                      syntax_error("A expected after H");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'L' )
                      syntax_error("L expected after HA");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'T' )
                      syntax_error("T expected after HAL");
                  return opcode_HALT;
                 break;
        case 'J': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'U' )
                      syntax_error("U expected after J");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'M' )
                      syntax_error("M expected after JU");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'P' )
                      syntax_error("P expected after JUM");
                  return opcode_JUMP;
                 break;
        case 'L': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'D' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'I' )
                          syntax_error("I expected after LD");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 == 'B' ) {
                          return opcode_LDIB;
                      } else if ( c2 == 'W' ) {
                          return opcode_LDIW;
                      } else
                          syntax_error("B or W expected after LDI");
                  } else if ( c2 == 'E' ) {
                      return DefCmpLE;
                  } else if ( c2 == 'O' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 == 'A' ) {
                          c2 = getc(in_f); putc(c2,out_f);
                          if ( c2 != 'D' )
                              syntax_error("D expected after LOA");
                          return opcode_LOAD;
                      } else if ( c2 == 'R' ) {
                          c2 = getc(in_f); putc(c2,out_f);
                          if ( c2 != '3' )
                              syntax_error("3 expected after LOR");
                          return opcode_LOR3;
                      } else
                          syntax_error("A or R expected after LO");
                  } else if ( c2 == 'T' ) {
                      return DefCmpLT;
                  } else
                      syntax_error("D or E or O or T expected after L");
                 break;
        case 'M': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'O' )
                      syntax_error("O expected after M");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'V' )
                      syntax_error("V expected after MO");
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == '2' ) {
                      return opcode_MOV2;
                  } else if ( c2 == 'R' ) {
                      return opcode_MOVR;
                  } else
                      syntax_error("2 or R expected after MOV");
                 break;
        case 'N': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'A' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'N' )
                          syntax_error("N expected after NA");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'D' )
                          syntax_error("D expected after NAN");
                      return opcode_NAND;
                  } else if ( c2 == 'E' ) {
                      return DefCmpNE;
                  } else if ( c2 == 'O' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'T' )
                          syntax_error("T expected after NO");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != '2' )
                          syntax_error("2 expected after NOT");
                      return opcode_NOT2;
                  } else
                      syntax_error("A or E or O expected after N");
                 break;
        case 'O': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 != 'V' )
                      syntax_error("V expected after O");
                  return DefCmpOV;
                 break;
        case 'P': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'C' ) {
                      return NoReg(4);
                  } else if ( c2 == 'O' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'P' )
                          syntax_error("P expected after PO");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'R' )
                          syntax_error("P expected after POP");
                      return opcode_POPR;
                  } else if ( c2 == 'U' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'S' )
                          syntax_error("S expected after PU");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'H' )
                          syntax_error("H expected after PUS");
                      return opcode_PUSH;
                  } else
                      syntax_error("C or O or U expected after P");
                 break;
        case 'R': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'e' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 't' )
                          syntax_error("t expected after Re");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'n' )
                          syntax_error("n expected after Ret");
                      return NoReg(7);
                  } else if ( c2 == 'E' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'T' )
                          syntax_error("T expected after RE");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'N' )
                          syntax_error("N expected after RET");
                      return opcode_RETN;
                  } else if ( c2 == 'O' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'T' )
                          syntax_error("T expected after RO");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'A' )
                          syntax_error("A expected after ROT");
                      return opcode_ROTA;
                  } else if ( c2 == '0' || c2 == '1' ) {
                      i = -100;
                      if ( c2 == '1' )
                          i -= 10;
                      c3 = getc(in_f); putc(c3,out_f);
                      if ( (c3 < '0') || (c3 > '9') )
                          syntax_error("Not a Register number!");
                      if ( (c2 == '1') && (c3 > '5') )
                          syntax_error("Not a valid Register number!");
                      i -= (c3 - '0');
                      return i;
                  } else
                      syntax_error("E or O or Register expected after R");
                 break;
        case 'S': putc(nc,out_f);
                  c2 = getc(in_f); putc(c2,out_f);
                  if ( c2 == 'C' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'M' )
                          syntax_error("M expected after SC");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'P' )
                          syntax_error("P expected after SCM");
                      return opcode_SCMP;
                  } else if ( c2 == 'H' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'F' )
                          syntax_error("F expected after SH");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'T' )
                          syntax_error("T expected after SHF");
                      return opcode_SHFT;
                  } else if ( c2 == 'P' ) {
                      return NoReg(5);
                  } else if ( c2 == 'T' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'O' )
                          syntax_error("O expected after ST");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'R' )
                          syntax_error("R expected after STO");
                      return opcode_STOR;
                  } else if ( c2 == 'U' ) {
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != 'B' )
                          syntax_error("B expected after SU");
                      c2 = getc(in_f); putc(c2,out_f);
                      if ( c2 != '3' )
                          syntax_error("3 expected after SUB");
                      return opcode_SUB3;
                  } else
                      syntax_error("C or H or P or T or U expected after S");
                 break;
        default : putc(nc,out_f);
                  fprintf(out_f,"\n??? LEXICAL ERROR !!!\n");
                exit(1);
      }
}

static void read_instruct(int f_lex, FILE * in_f, FILE * out_f) {
    int i, j;
    int cut, nxt;
    int op1, op2, op3;
    int vlist = 0;

    cut = f_lex;
    while ( cut == Number || cut == Identif ) {
        if ( (nxt = next_lex_term(in_f,out_f)) == EndLabel ) {
            if ( vlist )
              syntax_error("semicolon expected");
            if ( cut == Number ) {
                if ( (my_number < -VMR_RAM_Size) ||
                     (my_number >= (VMR_RAM_Size<<1)) ) {
                    fprintf(out_f,"\n*** ERROR: label value=%1d too high!\n\n",
                            my_number);
                    exit(1);
                  }
                my_number &= 0xFFFF;
                if ( my_number < VMR_RAM_Size ) {
                    fprintf(out_f,"\n*** ERROR: label value=%1d is too low!\n\n",
                            my_number);
                    exit(1);
                  }
                if ( cell_used[my_number-VMR_RAM_Size] ) {
                    fprintf(out_f,"\n*** ERROR: location %1d already used!\n\n",
                            my_number);
                    exit(1);
                  }
                cur_pc = my_number-VMR_RAM_Size;
            } else { /* if ( cut == Identif ) */
                for ( j = 0 ; j <= max_pc ; j++ )
                    if ( (Rlabels[j] != NULL) &&
                         (strcmp(Rlabels[j],my_string) == 0) ) {
                        fprintf(out_f,
                    "\n*** ERROR: label %s is defined more than once!\n\n",
                                my_string);
                        exit(1);
                      }
                if ( Rlabels[cur_pc] != NULL ) {
                    fprintf(out_f,
  "\n*** ERROR: another label (%s) is already defined for location %1d!\n\n",
                            my_string, cur_pc);
                    exit(1);
                  }
                Rlabels[cur_pc] = my_str_alloc(my_string);
                for ( j = 0 ; j < no_unid ; j++ )
                    if ( strcmp(unidentif[j].nam,my_string) == 0 ) {
                        for ( i = unidentif[j].fst; i >= 0 ; i = undef_link[i] ) {
                          int aux = undef_add[i];
                            aux += (cur_pc + VMR_RAM_Size);
                            if ( aux < 0 ) {
                                if ( (-aux) > ((undef_mask[i])>>1) ) {
                                    fprintf(out_f,
            "\n*** ERROR: label (%s) is too far away from location %1d!\n\n",
                                            my_string, i);
                                    exit(1);
                                  }
                            } else {
                                if ( (undef_mask[i] < 0xFFFF) &&
                                     (aux >= ((undef_mask[i])>>1)) ) {
                                    fprintf(out_f,
            "\n*** ERROR: label (%s) is too far away from location %1d!\n\n",
                                            my_string, i);
                                    exit(1);
                                  }
                              }
                            RAM[i] |= (aux & undef_mask[i]);
                          }
                        i = (no_unid-1);
                        if ( j < i )
                            unidentif[j] = unidentif[i];
                        no_unid = i;
                        break;
                      }
              }
            cut = next_lex_term(in_f,out_f);
        } else if ( (nxt == EndList) || (nxt == InstrSepar) ) {
            if ( cur_pc >= VMR_RAM_Size ) {
                fprintf(out_f,"\n*** ERROR: VMR RAM exhausted!\n\n");
                exit(1);
              }
            if ( cell_used[cur_pc] ) {
                fprintf(out_f,
                "\n*** ERROR: RAM cell number %d is already in use!\n\n",
                        cur_pc+VMR_RAM_Size );
                exit(1);
              }
            if ( cur_pc > max_pc )
                max_pc = cur_pc;
            cell_used[cur_pc] = TRUE;
            if ( cut == Number )
                op1 = my_number;
              else { /* if ( cut == Identif ) */
                int search = TRUE;
                  op1 = 0;
                  for ( j = 0 ; j <= max_pc ; j++ )
                      if ( (Rlabels[j] != NULL) &&
                           (strcmp(Rlabels[j],my_string) == 0) ) {
                          op1 = (j+VMR_RAM_Size);
                          search = FALSE;
                          break;
                        }
                  for ( j = 0 ; search && (j < no_ident) ; j++ )
                      if ( (strcmp(identifier[j].nam,my_string) == 0) ) {
                          op1 = identifier[j].val;
                          search = FALSE;
                        }
                  if ( search ) { /* add to to-do list */
                      for ( j = 0 ; j < no_unid ; j++ )
                          if ( strcmp(unidentif[j].nam,my_string) == 0 ) {
                              goto found_0;
                            }
                      /* not found */
                      j = no_unid;
                      unidentif[j].nam = my_str_alloc(my_string);
                      unidentif[j].fst = -1;
                      no_unid++;
                    found_0:
                      undef_link[cur_pc] = unidentif[j].fst;
                      unidentif[j].fst = cur_pc;
                      undef_mask[cur_pc] = 0xFFFF;
                      undef_add[cur_pc] = 0;
                      op1 = 0;
                    }
              }
            if ( (op1 > 0xFFFF) || (op1 < -32768) ) {
                fprintf(out_f,
            "\n*** ERROR: value %1d is too high to fit in location %1d!\n\n",
                        op1, cur_pc);
                exit(1);
              }
            RAM[cur_pc] = op1 & 0xFFFF;
            cur_pc++;
            if ( nxt == InstrSepar ) {
                vlist = 1;
                cut = next_lex_term(in_f,out_f);
            } else
                return;
        } else
            syntax_error("colon or semicolon or comma expected");
      }
    if ( cut == DefDecl ) {
        if ( (nxt = next_lex_term(in_f,out_f)) != Identif )
            syntax_error("Identifier expected");
        for ( j = 0 ; j <= max_pc ; j++ )
            if ( (Rlabels[j] != NULL) &&
                 (strcmp(Rlabels[j],my_string) == 0) ) {
                fprintf(out_f,
        "\n*** ERROR: label %s is already associated with location %1d !\n\n",
                        my_string, j);
                exit(1);
              }
        for ( j = 0 ; j < no_ident ; j++ )
            if ( (strcmp(identifier[j].nam,my_string) == 0) ) {
                fprintf(out_f,
                "\n*** ERROR: identifier %s is already declared !\n\n",
                        my_string);
                exit(1);
              }
        if ( no_ident >= MAX_DEF ) {
            fprintf(out_f,
                    "\n*** ERROR: more than %d identifiers defined !\n\n",
                    MAX_DEF);
            exit(1);
          }
        identifier[no_ident].nam = my_str_alloc(my_string);
        if ( (nxt = next_lex_term(in_f,out_f)) != Assign )
            syntax_error("equal sign expected");
        if ( (nxt = next_lex_term(in_f,out_f)) != Number )
            syntax_error("number expected");
        if ( (my_number < -VMR_RAM_Size) || (my_number >= (VMR_RAM_Size<<1)) )
            syntax_error("out of VM-R RAM address range");
        my_number &= 0xFFFF;
        identifier[no_ident].val = my_number;
        no_ident++;
        for ( j = 0 ; j < no_unid ; j++ )
            if ( strcmp(unidentif[j].nam,my_string) == 0 ) {
                for ( i = unidentif[j].fst; i >= 0 ; i = undef_link[i] ) {
                  int aux = undef_add[i];
                    RAM[i] |= (my_number & undef_mask[i]);
                  }
                i = (no_unid-1);
                if ( j < i )
                    unidentif[j] = unidentif[i];
                no_unid = i;
                break;
              }
        return;
      }
    if ( cur_pc >= VMR_RAM_Size ) {
        fprintf(out_f,"\n*** ERROR: VMR RAM exhausted!\n\n");
        exit(1);
      }
    if ( cell_used[cur_pc] ) {
        fprintf(out_f,"\n*** ERROR: RAM cell number %d is already in use!\n\n",
                cur_pc+VMR_RAM_Size );
        exit(1);
      }
    if ( cur_pc > max_pc )
        max_pc = cur_pc;
    cell_used[cur_pc] = TRUE;
    switch (cut) {
        case opcode_ADD3:
        case opcode_AND3:
        case opcode_LOR3:
        case opcode_LOAD:
        case opcode_SUB3:
        case opcode_NAND:
        case opcode_MOVR:
        case opcode_STOR:
            /* format F3 */
            nxt = next_lex_term(in_f,out_f);
            if ( ! IsReg(nxt) )
                syntax_error("Register expected");
            op1 = (NoReg(nxt))<<8;
            if ( (nxt = next_lex_term(in_f,out_f)) != InstrSepar )
                syntax_error("comma expected");
            nxt = next_lex_term(in_f,out_f);
            if ( ! IsReg(nxt) )
                syntax_error("Register expected");
            op2 = (NoReg(nxt))<<4;
            if ( (nxt = next_lex_term(in_f,out_f)) != InstrSepar )
                syntax_error("comma expected");
            nxt = next_lex_term(in_f,out_f);
            if ( ! IsReg(nxt) )
                syntax_error("Register expected");
            op3 = NoReg(nxt);
            RAM[cur_pc] = cut | op1 | op2 | op3;
          break;
        case opcode_ROTA:
        case opcode_PUSH:
        case opcode_SHFT:
        case opcode_POPR:
            /* format F2 */
            nxt = next_lex_term(in_f,out_f);
            if ( ! IsReg(nxt) )
                syntax_error("Register expected");
            op1 = (NoReg(nxt))<<8;
            if ( (nxt = next_lex_term(in_f,out_f)) != InstrSepar )
                syntax_error("comma expected");
            nxt = next_lex_term(in_f,out_f);
            if ( ! IsReg(nxt) )
                syntax_error("Register expected");
            op2 = (NoReg(nxt))<<4;
            if ( (nxt = next_lex_term(in_f,out_f)) != InstrSepar )
                syntax_error("comma expected");
            nxt = next_lex_term(in_f,out_f);
            if ( nxt != Number  )
                syntax_error("Number expected");
            op3 = my_number;
            if ( (op3 > 16) || (op3 < -8) )
                syntax_error("4 bit number expected");
            if ( (cut == opcode_PUSH) || (cut == opcode_POPR) ) {
                if ( op3 < 1 )
                    syntax_error("strictly positive number expected");
                op3--;
              } else {
                if ( op3 == 0 )
                    syntax_error("non-null number expected");
                if ( op3 > 8 )
                    syntax_error("number < 9 expected");
                if ( op3 < 0 )
                    op3 = 8 | (-1-my_number);
                  else
                    op3 = my_number -1;
              }
            RAM[cur_pc] = cut | op1 | op2 | op3;
          break;
        case opcode_ADD1:
        case opcode_AND1:
        case opcode_LDIB:
            /* format F1 */
            nxt = next_lex_term(in_f,out_f);
            if ( ! IsReg(nxt) )
                syntax_error("Register expected");
            op1 = (NoReg(nxt))<<8;
            if ( (nxt = next_lex_term(in_f,out_f)) != InstrSepar )
                syntax_error("comma expected");
            nxt = next_lex_term(in_f,out_f);
            if ( nxt != Number  )
                syntax_error("Number expected");
            op2 = my_number;
            if ( (op2 > 255) || (op2 < -128) )
                syntax_error("8 bit number expected");
            if ( cut == opcode_AND1 ) {
                if ( op2 < 0 )
                    syntax_error("positive number expected");
              } else {
                op2 &= 0xFF;
              }
            RAM[cur_pc] = cut | op1 | op2;
          break;
        case opcode_CJMP:
            /* format FJ */
            nxt = next_lex_term(in_f,out_f);
            switch ( nxt ) {
              case DefCmpAL:
                  op1 = 0x000;
                break;
              case DefCmpEQ:
                  op1 = 0x200;
                break;
              case DefCmpGT:
                  op1 = 0x400;
                break;
              case DefCmpGE:
                  op1 = 0x600;
                break;
              case DefCmpOV:
                  op1 = 0x800;
                break;
              case DefCmpNE:
                  op1 = 0xA00;
                break;
              case DefCmpLE:
                  op1 = 0xC00;
                break;
              case DefCmpLT:
                  op1 = 0xE00;
                break;
              default:
                syntax_error("Jump condition expected");
            }
        case opcode_JUMP:
            nxt = next_lex_term(in_f,out_f);
            if ( nxt == Identif  ) {
              int search = TRUE;
                op2 = 0;
                for ( j = 0 ; j <= max_pc ; j++ )
                    if ( (Rlabels[j] != NULL) &&
                         (strcmp(Rlabels[j],my_string) == 0) ) {
                        op2 = j - (cur_pc+1);
                        search = FALSE;
                        break;
                      }
                for ( j = 0 ; search && (j < no_ident) ; j++ )
                    if ( (strcmp(identifier[j].nam,my_string) == 0) ) {
                        op2 = (identifier[j].val) - (cur_pc+1+VMR_RAM_Size);
                        search = FALSE;
                      }
                if ( search ) { /* add to to-do list */
                    for ( j = 0 ; j < no_unid ; j++ )
                        if ( strcmp(unidentif[j].nam,my_string) == 0 ) {
                            goto found_1;
                          }
                    /* not found */
                    j = no_unid;
                    unidentif[j].nam = my_str_alloc(my_string);
                    unidentif[j].fst = -1;
                    no_unid++;
                  found_1:
                    undef_link[cur_pc] = unidentif[j].fst;
                    unidentif[j].fst = cur_pc;
                    undef_mask[cur_pc] = 0x1FF;
                    undef_add[cur_pc] = - (cur_pc+1+VMR_RAM_Size);
                    op2 = 0;
                  }
            } else if ( nxt == Number  )
                op2 = my_number;
              else
                syntax_error("Offset or Label expected");
            if ( (op2 > 511) || (op2 < -256) )
                syntax_error("9 bit number expected");
            op2 &= 0x1FF;
            if ( cut == opcode_CJMP )
                RAM[cur_pc] = cut | op1 | op2;
              else
                RAM[cur_pc] = (cut | op2)&0xFFFF;
          break;
        case opcode_MOV2:
        case opcode_NOT2:
        case opcode_SCMP:
            /* pseudo 2 registers */
            nxt = next_lex_term(in_f,out_f);
            if ( ! IsReg(nxt) )
                syntax_error("Register expected");
            if ( cut == opcode_SCMP )
                op1 = (NoReg(nxt))<<4;
              else
                op1 = (NoReg(nxt))<<8;
            if ( (nxt = next_lex_term(in_f,out_f)) != InstrSepar )
                syntax_error("comma expected");
            nxt = next_lex_term(in_f,out_f);
            if ( ! IsReg(nxt) )
                syntax_error("Register expected");
            if ( cut == opcode_SCMP )
                op2 = NoReg(nxt);
              else {
                op2 = (NoReg(nxt))<<4;
                if ( cut == opcode_NOT2 )
                    op2 |= NoReg(nxt);
              }
            RAM[cur_pc] = (cut | op1 | op2)&0xFFFF;
          break;
        case opcode_LDIW:
            /* pseudo 1 register + 1 word */
            nxt = next_lex_term(in_f,out_f);
            if ( ! IsReg(nxt) )
                syntax_error("Register expected");
            op1 = (NoReg(nxt))<<8;
            RAM[cur_pc] = (cut | op1)&0xFFFF;
            cur_pc++;
            if ( cur_pc > max_pc )
                max_pc = cur_pc;
            cell_used[cur_pc] = TRUE;
            if ( (nxt = next_lex_term(in_f,out_f)) != InstrSepar )
                syntax_error("comma expected");
            nxt = next_lex_term(in_f,out_f);
            if ( nxt == Identif  ) {
              int search = TRUE;
                op2 = 0;
                for ( j = 0 ; j <= max_pc ; j++ )
                    if ( (Rlabels[j] != NULL) &&
                         (strcmp(Rlabels[j],my_string) == 0) ) {
                        op2 = (j+VMR_RAM_Size);
                        search = FALSE;
                        break;
                      }
                for ( j = 0 ; search && (j < no_ident) ; j++ )
                    if ( (strcmp(identifier[j].nam,my_string) == 0) ) {
                        op2 = identifier[j].val;
                        search = FALSE;
                      }
                if ( search ) { /* add to to-do list */
                    for ( j = 0 ; j < no_unid ; j++ )
                        if ( strcmp(unidentif[j].nam,my_string) == 0 ) {
                            goto found_2;
                          }
                    /* not found */
                    j = no_unid;
                    unidentif[j].nam = my_str_alloc(my_string);
                    unidentif[j].fst = -1;
                    no_unid++;
                  found_2:
                    undef_link[cur_pc] = unidentif[j].fst;
                    unidentif[j].fst = cur_pc;
                    undef_mask[cur_pc] = 0xFFFF;
                    undef_add[cur_pc] = 0;
                    op2 = 0;
                  }
            } else if ( nxt == Number  )
                op2 = my_number;
              else
                syntax_error("Number or Identifier expected");
            if ( (op2 > 0xFFFF) || (op2 < -32768) )
                syntax_error("16 bit number expected");
            op2 &= 0xFFFF;
            RAM[cur_pc] = op2;
          break;
        case opcode_CALL:
            /* pseudo 1 register */
            nxt = next_lex_term(in_f,out_f);
            if ( ! IsReg(nxt) )
                syntax_error("Register expected");
            op1 = NoReg(nxt);
            RAM[cur_pc] = (cut | op1)&0xFFFF;
          break;
        case opcode_HALT:
        case opcode_RETN:
            /* pseudo no register */
            RAM[cur_pc] = cut & 0xFFFF;
          break;

        default:
            syntax_error("instruction or number or DEFL expected");
        }
    cur_pc++;
}


int main(int argc, char ** argv) {
    char inp_f_nam[MAX_CHAR];
    char out_f_nam[MAX_CHAR];
    FILE * infp, * oufp;
    int nxt, j;

    printf(
"VM-R Assembler, Copyright C 2001 by Giovanni Chiola <chiola@disi.unige.it>\n"
            );
    if ( argc > 2 ) {
      sprintf(inp_f_nam,"%s",argv[1]);
      sprintf(out_f_nam,"%s",argv[2]);
    }
    else
      if ( argc == 2 ) {
        sprintf(inp_f_nam,"%s",argv[1]);
        printf("Output file name:"); fflush(stdout);
        scanf("%s",out_f_nam);
      }
      else {
        printf("Input file name:"); fflush(stdout);
        scanf("%s",inp_f_nam);
        printf("Output file name:"); fflush(stdout);
        scanf("%s",out_f_nam);
      }
    infp = fopen(inp_f_nam,"r");
    if ( infp == NULL ) {
        printf("\n*** ERROR: couldnt open input file %s\n\n",inp_f_nam);
        exit(1);
      }
    oufp = fopen(out_f_nam,"w");
    if ( oufp == NULL ) {
        printf("\n*** ERROR: couldnt open output file %s\n\n",out_f_nam);
        exit(1);
      }
    init();
    for ( nxt = next_lex_term(infp,stdout) ; nxt != EndProgr ;
          nxt = next_lex_term(infp,stdout) )
        read_instruct(nxt,infp,stdout);
    if ( no_unid > 0 ) {
        printf("\n*** ERROR: the following identifiers are undefined:\n");
        while ( no_unid > 0 ) {
            no_unid--;
            printf("  %s (used in the following locations:",
                  unidentif[no_unid].nam);
            for ( j = unidentif[no_unid].fst; j >= 0 ; j = undef_link[j] ) {
                printf(" %1d", j);
              }
            printf("\n");
          }
        exit(1);
      }
    for ( j = 0 ; j <= max_pc ; j++ )
        fprintf(oufp,"%4x\n",RAM[j]);
    fprintf(oufp,"\n");
}

